/* Tiny css compression
 * 2006 Viliam Holub
 */


#include <errno.h>
#include <stdio.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <stdbool.h>


const char *const version = "Css lossless compression  Version 0.2 Viliam Holub\n";
const char *const usage = "usage: cssc input_file output_file\n";
const char *const error_msg = "Invalid number of parameters\n";

/* Input/output files */
FILE *fi = NULL;
FILE *fo = NULL;


bool
error( bool fatal, const char *f, ...)

{
	va_list ap;

	fflush( stdout);
	fprintf( stderr, "cssc: ");

	va_start( ap, f);
	vfprintf( stderr, f, ap);
	va_end( ap);

	if (fatal)
		exit( 1);
	return false;
}


bool
cbreak( int c)

{
	return c == '{' || c == '}' || c == ',' || c == ':' || c == ';'
		|| c == '(' || c == ')' || c == '>' || c == '+';
}


int
put( int c)

{
	return fputc( c, fo);
}


void
put_space( void)

{
	put( ' ');
}


void
copy_string( void)

{
	bool back = true;
	int c;

	put( '"');
	while (c = fgetc( fi), c != EOF)
	{
		if (c == '\\')
			back != true;
		else
			back = false;
		if (c == '"' && !back)
			break;
		put( c);
	}
	put( '"');
}


bool
hex( int c)

{
	return (c >= '0' && c <= '9')
		|| (c >= 'a' && c <= 'f')
		|| (c >= 'A' && c <= 'F');
}


int
tolow( int c)
	
{
	return (c >= 'A' && c <= 'Z')
		? c-'A'+'a' : c;
}


void
color( void)

{
	int i=0, c, d;
	char buf[ 6], *b;

	put( '#');
	while (c = fgetc( fi), c != EOF)
	{
		if (!hex( c))
		{
			ungetc( c, fi);
			break;
		}
		buf[ i++] = tolow( c);
		if (i == 6)
			break;
	}

	if (i == 6 && buf[ 0] == buf[ 1] && buf[ 2] == buf[ 3]
			&& buf[ 4] == buf[ 5])
		{
			buf[ 1] = buf[ 2];
			buf[ 2] = buf[ 4];
			i = 3;
		}

	for (d=0; d!=i; d++)
		put( buf[ d]);
}


int
main( int argc, char **argv)

{

	/* Check parameters
	 */
	if (argc <= 2 || argc >= 4)
		error( true, "%s%s%s",
				argc <= 1 ? version : "",
				argc == 2 || argc >= 4 ? error_msg : "",
				usage);

	/* Open files
	 */
	fi = fopen( argv[ 1], "r");
	if (!fi)
		error( true, "%s: %s\n", argv[ 1], strerror( errno));
	else
	{
		fo = fopen( argv[ 2], "w");
		if (!fo)
			error( false, "%s: %s\n", argv[ 2], strerror( errno));
	}

	if (fi && fo)
	{
		int c;
		int ostate;
		bool semi = false;
		int state = 0;
		int space = 0; /* -1 = nospace hard, 0 = nospace, 1 = space */

		errno = 0;
		while (c = fgetc( fi), c != EOF)
		{
			switch (state)
			{
				case 0:
					if (c == '/')
						state = 1;
					break;
				case 1:
					if (c == '*')
						state = 2;
					else
					{
						space = !!isspace( c);
						state = 4;
					}
					break;
				case 2:
					if (c == '*')
						state = 3;
					break;
				case 3:
					if (c == '/')
					{
						state = 4;
						space = -1;
					}
					else if (c != '*')
						state = 2;
					break;
				/* Main tag cycle */
				case 4:
					if (isspace( c))
					{
						c = EOF;
						space = space == -1 ? space : 1;
					}
					else if (c == '\\')
					{
						if (space == 1)
							put_space();
						space = 0;
						ostate = 4;
						state = 8;
					}
					else if (cbreak( c))
					{
						if (c == '{')
							state = 9;
						space = -1;
					}
					else if (c == '/')
					{ c = EOF; ostate = 4; state = 5; }
					else if (c == '"')
					{
						if (space == 1)
							put_space();
						copy_string();
						space = -1;
						c = EOF;
					}
					else
					{
						if (space == 1)
							put_space();
						space = 0;
					}
					break;
				/* Inline comment */
				case 5:
					if (c == '*')
					{ c = EOF; state = 6; }
					else
					{
						if (space == 1)
							put_space();
						put( '/');
						space = !!isspace( c);
						state = ostate;
					}
					break;
				case 6:
					if (c == '*')
						state = 7;
					c = EOF;
					break;
				case 7:
					if (c == '*')
						;
					if (c == '/')
					{
						state = ostate;
						space = space == -1 ? space : 1;
					}
					else
						state = 6;
					c = EOF;
					break;
				/* Backslash skip */
				case 8:
					state = ostate;
					break;
				/* Tag cycle */
				case 9:
					if (isspace( c))
					{
						c = EOF;
						space = space == -1 ? space : 1;
					}
					else
					{
						if (semi)
						{
							semi = false;
							if (c != '}')
							{
								put( ';');
							}
							else
								state = 4;
							space = -1;
						}
						
						if (c == ';')
						{
							semi = true;
							c = EOF;
						}
						else if (c == '#')
						{
							color();
							space = 1;
							c = EOF;
						}
						else if (c == '\\')
						{
							if (space == 1)
								put_space();
							space = 0;
							ostate = 9;
							state = 8;
						}
						else if (cbreak( c))
						{
							if (c == '}')
								state = 4;
							space = -1;
						}
						else if (c == '/')
						{ c = EOF; ostate = 9; state = 5; }
						else if (c == '"')
						{
							if (space == 1)
								put_space();
							copy_string();
							space = -1;
							c = EOF;
						}
						else
						{
							if (space == 1)
								put_space();
							space = false;
						}
					}
					break;
			}

			if (c != EOF)
				put( c);
		}

		if (!feof( fi))
			error( false, "%s: %s\n", argv[ 1], strerror( errno));
		if (ferror( fo) != 0)
			error( false, "%s: %s\n", argv[ 1], strerror( errno));
	}

	/* Close files
	 */
	if (fclose( fi))
		error( false, "%s: %s\n", argv[ 1], strerror( errno));
	if (fo && fclose( fo))
		error( false, "%s: %s\n", argv[ 2], strerror( errno));

	return 0;
}
